#pragma once
#include <stdint.h>
#include <stdbool.h>
#include <string.h>


// P2 addpins/addbits macro
#define addpins(a,b) a + ((b)<<6)
#define addbits(a,b) a + ((b)<<5)

#ifdef __propeller2__
#define longset __builtin_longset
#else
static inline void *longset(uint32_t *dst, uint32_t val,size_t len) {
    uint32_t *d = dst;
    do {
        *d++ = val;
    } while(--len);
    return dst;
}
#endif

static inline uint16_t rgbsqz(uint32_t x) {
#ifdef __bropeller2__
    __asm {
        rgbsqz x
    }
    return x;
#else
    return ((x>>11)&0x001F) | ((x>>13)&0x07E0) | ((x>>16)&0xF800);
#endif
}

static inline uint32_t rgbexp(uint16_t x) {
#ifdef __propeller2__
    uint32_t tmp = x;
    __asm {
        rgbexp tmp
    }
    return tmp;
#else
    return ((x&0x001F)<<11)|((x&0x001C)<<8)|((x&0x07E0)<<13)|((x&0x0600)<<7)|((x&0xF800)<<16)|((x&0xE000)<<11);
#endif
}

static inline unsigned incmod(unsigned a,unsigned b) {
    if (a == b) return 0;
    return a+1;
}
static inline unsigned decmod(unsigned a,unsigned b) {
    if (a == 0) return b;
    return a-1;
}

static inline int32_t fixdivS32U32(int32_t a,uint32_t b) {
    return ((int64_t)a<<16)/b;
}


static inline int32_t fixmulS32S32(int32_t a,int32_t b) {
    return ((int64_t)a*(int64_t)b)>>16;
}

static inline int32_t fixmulS32U16(int32_t a,int32_t b) {
    return fixmulS32S32(a,b&0xFFFF);
}

static inline uint32_t saturate_u8(uint32_t x) {
    return x > 255 ? 255 : x;
}
static inline uint32_t saturate_u16(uint32_t x) {
    return x > 65535 ? 65535 : x;
}

static inline uint32_t addpix(uint32_t a,uint32_t b) {
    #ifdef __bropeller2__
    __asm {
        addpix a,b
    }
    return a;
#else
    return (
        (saturate_u8(((a>> 0)&255)+((b>> 0)&255)) <<  0) +
        (saturate_u8(((a>> 8)&255)+((b>> 8)&255)) <<  8) +
        (saturate_u8(((a>>16)&255)+((b>>16)&255)) << 16) +
        (saturate_u8(((a>>24)&255)+((b>>24)&255)) << 24)
    );
#endif
}

static inline uint32_t blnpix(uint32_t a,uint32_t b,uint8_t piv) {
    #ifdef __propeller2__
    __asm {
        setpiv piv
        blnpix a,b
    }
    return a;
#else
    uint8_t notpiv = piv^255;
    return (
        (saturate_u8((((a>> 0)&255)*notpiv+((b>> 0)&255)*piv+255)>>8) <<  0) +
        (saturate_u8((((a>> 8)&255)*notpiv+((b>> 8)&255)*piv+255)>>8) <<  8) +
        (saturate_u8((((a>>16)&255)*notpiv+((b>>16)&255)*piv+255)>>8) << 16) +
        (saturate_u8((((a>>24)&255)*notpiv+((b>>24)&255)*piv+255)>>8) << 24)
    );
#endif
}


struct vec2 {
    int32_t x,y;
};
struct vec3 {
    int32_t x,y,z;
};
struct matrix {
    int32_t xx,xy,xz,xt;
    int32_t yx,yy,yz,yt;
    int32_t zx,zy,zz,zt;
};

static inline struct vec2 sincos32(uint32_t angle) {
    #ifdef __propeller2__
        int32_t x,y;
        __asm {
            qrotate ##0x10000,angle
            getqx x
            getqy y
        };
        return (struct vec2){x,y};
    #else
        double a = (double)angle * 1.4629180792671596e-09;
        return (struct vec2){(int32_t)(cos(a)*0x10000),(int32_t)(sin(a)*0x10000)};
    #endif
}



// PSRAM type values
#define EXMEM_PSRAM16 1
#define EXMEM_PSRAM8 2
#define EXMEM_PSRAM4 3
#define EXMEM_HYPERRAM 4


