#include <stdint.h>
#include <stdio.h>
#include "hal.h"
#include "hal_scancodes.h"

#include "assets.h"

uint16_t screen[320*240];

uint32_t dither_mtx[2][2] = {
    {0x00000000,0x04020400},
    {0x06030600,0x02010200}
};



void drawSpanTest(int x, int length, int y, uint32_t color) {
    if (y < 0 || y >= 240) return;
    if (x < 0 || x >= 320) return;
    if (length <= 0 || x+length < 0 || x+length >= 321) return;
    uint16_t *base = &screen[x+y*320];
    do {
        //*base ^= 0xFFFF;
        *base = rgbsqz(addpix(rgbexp(*base),color));
        base++;
    } while (--length);
}

void drawPolygonTest(struct vec2 *verts ,unsigned vt_count, uint32_t color) {
    if (vt_count < 3) return;
    
    // Setup top vertex
    int top_vertex = 0;
    int32_t cur_y = verts[0].y;
    for (unsigned i=1;i<vt_count;i++) {
        int32_t vy = verts[i].y;
        if (verts[i].y <= cur_y) {
            top_vertex = i;
            cur_y = vy;
        }
    }

    int left_vertex = top_vertex, right_vertex = top_vertex;
    int32_t left_slope = 0, right_slope = 0;
    int32_t left_x = verts[top_vertex].x;
    int32_t right_x = left_x;

    vt_count--;
    int vt_left = vt_count;

    //printf("top_vertex %d\n",top_vertex);
    
    for (;;) {
        // load left vertices
        next_vert:
        if (vt_left < 0) break;
        if (cur_y == verts[left_vertex].y) {
            do {
                left_x = verts[left_vertex].x;
                if (--vt_left < 0) {
                    left_vertex = right_vertex; // bottom vertex
                    break;
                } else {
                    left_vertex = decmod(left_vertex,vt_count);
                }
                //printf("left_vertex %d\n",left_vertex);
            } while (cur_y == verts[left_vertex].y);
            if (cur_y == verts[left_vertex].y) break; // stop divide by zero exception? 
            left_slope = fixdivS32U32(verts[left_vertex].x - left_x, verts[left_vertex].y - cur_y);
        }
        if (cur_y == verts[right_vertex].y) {
            do {
                right_x = verts[right_vertex].x;
                if (--vt_left < 0) {
                    right_vertex = left_vertex; // bottom vertex
                    break;
                } else {
                    right_vertex = incmod(right_vertex,vt_count);
                }
                //printf("right_vertex %d\n",right_vertex);
            } while (cur_y == verts[right_vertex].y);
            right_slope = fixdivS32U32(verts[right_vertex].x - right_x, verts[right_vertex].y - cur_y);
        }

        bool next_left = verts[left_vertex].y <= verts[right_vertex].y;
        int32_t trap_bottom = next_left ? verts[left_vertex].y : verts[right_vertex].y;

        // Handle advance to next scanline
        if (cur_y & 0xFFFF) {
            if (((cur_y+0x10000)&~0xFFFF) > trap_bottom) { // no spans to render, actually
                // TODO: this should be s32*u16 multiply
                if (next_left) right_x += fixmulS32U16(right_slope,(trap_bottom-cur_y));
                else           left_x  += fixmulS32U16(left_slope ,(trap_bottom-cur_y));
                cur_y = trap_bottom;
                continue;
            }
            int32_t to_next = 0x10000 - (cur_y&0xFFFF);
            right_x += fixmulS32U16(right_slope,to_next);
            left_x  += fixmulS32U16(left_slope ,to_next);
            cur_y += to_next;
        }
        if (cur_y != trap_bottom) { // Need this if to not draw bottom edges
            // Draw spans
            for(;;) {
                int32_t left_int = (left_x+0xFFFF)>>16, right_int = (right_x+0xFFFF)>>16;
                //printf("cur_y: %08X, left_x: %08X, right_x :%08X, left_int: %d right_int: %d\n",cur_y,left_x,right_x,left_int,right_int);
                drawSpanTest(left_int,right_int - left_int,cur_y>>16,color);

                if (cur_y + 0x10000 <= trap_bottom) {
                    cur_y += 0x10000;
                    left_x  += left_slope;
                    right_x += right_slope;
                    if (cur_y == trap_bottom) goto next_vert;
                } else {
                    break;
                }
            }
        }

        // handle advance to next vertex
        if (!next_left) left_x  += fixmulS32U16(left_slope ,trap_bottom - cur_y);
        else            right_x += fixmulS32U16(right_slope,trap_bottom - cur_y);
        cur_y = trap_bottom;
    }
}

static void mt_identity(struct matrix *mt) {
    *mt = (struct matrix){
        1<<16,    0,    0,    0,
            0,1<<16,    0,    0,
            0,    0,1<<16,    0
    };
}

static void mt_translate(struct matrix *mt, int32_t x, int32_t y, int32_t z) {
    mt->xt += x;
    mt->yt += y;
    mt->zt += z;
}

static void mt_rotateX(struct matrix *mt, uint32_t angle) {
    struct vec2 sincos = sincos32(angle); // X is cos, Y is sin
    struct matrix tmp = *mt;

    // X stays the same

    // This all could be made faster in many ways
    mt->yx = fixmulS32S32(sincos.x,tmp.yx) + fixmulS32S32(-sincos.y,tmp.zx);
    mt->yy = fixmulS32S32(sincos.x,tmp.yy) + fixmulS32S32(-sincos.y,tmp.zy);
    mt->yz = fixmulS32S32(sincos.x,tmp.yz) + fixmulS32S32(-sincos.y,tmp.zz);
    mt->yt = fixmulS32S32(sincos.x,tmp.yt) + fixmulS32S32(-sincos.y,tmp.zt);

    mt->zx = fixmulS32S32(sincos.y,tmp.yx) + fixmulS32S32(sincos.x,tmp.zx);
    mt->zy = fixmulS32S32(sincos.y,tmp.yy) + fixmulS32S32(sincos.x,tmp.zy);
    mt->zz = fixmulS32S32(sincos.y,tmp.yz) + fixmulS32S32(sincos.x,tmp.zz);
    mt->zt = fixmulS32S32(sincos.y,tmp.yt) + fixmulS32S32(sincos.x,tmp.zt);

}

static void mt_rotateY(struct matrix *mt, uint32_t angle) {
    struct vec2 sincos = sincos32(angle); // X is cos, Y is sin
    struct matrix tmp = *mt;


    // This all could be made faster in many ways
    mt->xx = fixmulS32S32(sincos.x,tmp.xx) + fixmulS32S32(sincos.y,tmp.zx);
    mt->xy = fixmulS32S32(sincos.x,tmp.xy) + fixmulS32S32(sincos.y,tmp.zy);
    mt->xz = fixmulS32S32(sincos.x,tmp.xz) + fixmulS32S32(sincos.y,tmp.zz);
    mt->xt = fixmulS32S32(sincos.x,tmp.xt) + fixmulS32S32(sincos.y,tmp.zt);

    // Y stays the same

    mt->zx = fixmulS32S32(-sincos.y,tmp.xx) + fixmulS32S32(sincos.x,tmp.zx);
    mt->zy = fixmulS32S32(-sincos.y,tmp.xy) + fixmulS32S32(sincos.x,tmp.zy);
    mt->zz = fixmulS32S32(-sincos.y,tmp.xz) + fixmulS32S32(sincos.x,tmp.zz);
    mt->zt = fixmulS32S32(-sincos.y,tmp.xt) + fixmulS32S32(sincos.x,tmp.zt);

}

static void mt_rotateZ(struct matrix *mt, uint32_t angle) {
    struct vec2 sincos = sincos32(angle); // X is cos, Y is sin
    struct matrix tmp = *mt;


    // This all could be made faster in many ways
    mt->xx = fixmulS32S32(sincos.x,tmp.xx) + fixmulS32S32(-sincos.y,tmp.yx);
    mt->xy = fixmulS32S32(sincos.x,tmp.xy) + fixmulS32S32(-sincos.y,tmp.yy);
    mt->xz = fixmulS32S32(sincos.x,tmp.xz) + fixmulS32S32(-sincos.y,tmp.yz);
    mt->xt = fixmulS32S32(sincos.x,tmp.xt) + fixmulS32S32(-sincos.y,tmp.yt);

    mt->yx = fixmulS32S32(sincos.y,tmp.xx) + fixmulS32S32(sincos.x,tmp.yx);
    mt->yy = fixmulS32S32(sincos.y,tmp.xy) + fixmulS32S32(sincos.x,tmp.yy);
    mt->yz = fixmulS32S32(sincos.y,tmp.xz) + fixmulS32S32(sincos.x,tmp.yz);
    mt->yt = fixmulS32S32(sincos.y,tmp.xt) + fixmulS32S32(sincos.x,tmp.yt);

    // Z stays the same
}

static struct vec3 vt_transform32(struct vec3 in_vt,bool position,struct matrix *matrix) {
    struct vec3 out_vt = {0,0,0};
    out_vt.x += fixmulS32S32(in_vt.x,matrix->xx);
    out_vt.x += fixmulS32S32(in_vt.y,matrix->xy);
    out_vt.x += fixmulS32S32(in_vt.z,matrix->xz);

    out_vt.y += fixmulS32S32(in_vt.x,matrix->yx);
    out_vt.y += fixmulS32S32(in_vt.y,matrix->yy);
    out_vt.y += fixmulS32S32(in_vt.z,matrix->yz);

    out_vt.z += fixmulS32S32(in_vt.x,matrix->zx);
    out_vt.z += fixmulS32S32(in_vt.y,matrix->zy);
    out_vt.z += fixmulS32S32(in_vt.z,matrix->zz);

    if (position) {
        out_vt.x += matrix->xt;
        out_vt.y += matrix->yt;
        out_vt.z += matrix->zt;
    }
    return out_vt;
}

static struct vec2 vt_project(struct vec3 in_vt) {
    // Note: bad things about to happen if Z is zero/negative
    // Note 2: rewrite, this looses loads of precision
    struct vec2 out_vt;
    out_vt.x = fixdivS32U32(in_vt.x,in_vt.z)*320 + (160<<16);
    out_vt.y = fixdivS32U32(in_vt.y,in_vt.z)*320 + (120<<16);
    return out_vt;
}

struct vec3 cube_verts[8] = {
    {-1<<16,-1<<16,-1<<16},
    {-1<<16,-1<<16,+1<<16},
    {-1<<16,+1<<16,-1<<16},
    {-1<<16,+1<<16,+1<<16},
    {+1<<16,-1<<16,-1<<16},
    {+1<<16,-1<<16,+1<<16},
    {+1<<16,+1<<16,-1<<16},
    {+1<<16,+1<<16,+1<<16},
};

uint8_t cube_faces[] = {
    4,      0, 4, 6, 2, // front face
    4,      3, 7, 5, 1, // back face
    4,      0, 1, 5, 4, // top face
    4,      6, 7, 3, 2, // bottom face
    4,      1, 0, 2, 3, // left face
    4,      7, 6, 4, 5, // right face
    0
};

static struct matrix cur_mtx;


void draw_model(struct vec3 *verts, unsigned vt_count, uint8_t *faces, unsigned max_ngon) {
    struct vec2 *xf_buffer = __builtin_alloca(vt_count*sizeof(struct vec2));
    struct vec2 *face_buffer = __builtin_alloca(max_ngon*sizeof(struct vec2));

    for (unsigned i=0;i<vt_count;i++) {
        struct vec3 tmp = vt_transform32(verts[i],true,&cur_mtx);
        xf_buffer[i] = vt_project(tmp);
        //printf("XF i: %d xf.x: %08X xf.y: %08X tmp.x: %08X tmp.y: %08X tmp.z: %08X  \n",i,xf_buffer[i].x,xf_buffer[i].y,tmp.x,tmp.y,tmp.z);
    }

    uint32_t color = 0xFFF00000;
    uint8_t *faces_orig = faces;

    for(;;) {
        unsigned face_vts = *faces++;
        if (face_vts == 0) break;
        //printf("Prmitive with %d verts\n",face_vts);
        for (unsigned i=0;i<face_vts;i++) {
            //printf("%d\n",faces-faces_orig);
            uint8_t vt_id = *faces++; // Have to this up here because flexspin shits the bed when you auto-increment when moving a struct
            face_buffer[i] = xf_buffer[vt_id];
            //printf("GS i: %d xf.x: %08X xf.y: %08X\n",i,face_buffer[i].x,face_buffer[i].y);
        }
        drawPolygonTest(face_buffer,face_vts,color);
        color = ((color << 5) | (color>>(32-5))) ^ 0x1111;
    }

}

int main(int argc, char **argv) {
    HAL_startup();
    printf("Hello!\n");

    int frame = 0;

    int32_t eyeX=0,eyeY=0,eyeZ=-10<<16;

    for(;;) {
        HAL_poll_inputs();

        if (HAL_keystate(KEY_SPACE)) eyeY -= 0x700;
        if (HAL_keystate(KEY_LSHIFT)) eyeY += 0x700;
        if (HAL_keystate(KEY_W)) eyeZ += 0x700;
        if (HAL_keystate(KEY_S)) eyeZ -= 0x700;
        if (HAL_keystate(KEY_A)) eyeX -= 0x700;
        if (HAL_keystate(KEY_D)) eyeX += 0x700;


        //if (HAL_keystate(KEY_DOWN)) testy+= 0x300;
        //if (HAL_keystate(KEY_LEFT)) testx-= 0x300;
        //if (HAL_keystate(KEY_RIGHT)) testx+= 0x300;

        //memcpy(screen,remmu,320*240*2);
        {
            uint16_t *src = remmu, *dst = screen;
            for (int y=0;y<240;y++) {
                //uint32_t color = blnpix(0x20202000u,0x60606000u,(y*(65535/240))>>8);
                //uint32_t fill = rgbsqz(addpix(color,dither_mtx[y&1][0]))+ (rgbsqz(addpix(color,dither_mtx[y&1][1]))<<16);
                //longset((uint32_t*)(screen+320*y+0),fill,160);
                int piv = ((y*((128<<8)/240))>>8)+64;
                //for(int x=0;x<320;x++) *dst++ = rgbsqz(blnpix(rgbexp(*src++),0,piv)+dither_mtx[y&1][x&1]);
                for (int x=0;x<320;x+=2,dst+=2,src+=2) *(uint32_t*)dst = (*(uint32_t*)src>>1)&0x7BEF7BEF;
            }
        }

        for (int i=0;i<240;i++) {screen[i*320] = 0xFFFF;screen[i*320+319] = 0xFFFF;}
        for (int i=0;i<320;i++) {screen[i] = 0xFFFF;screen[i+239*320] = 0xFFFF;}

        screen[320+(frame&255)] = 0x0000;

        //drawSpanTest(10,100,80,0x11111100);

        /*
        struct vec2 test[3] = {
            {100<<16,50<<16},
            {180<<16,200<<16},
            {80<<16,100<<16},
        };
        drawPolygonTest(test,3,0x00400000);
        struct vec2 test2[4] = {
            {100<<16,50<<16},
            {200<<16,45<<16},
            {300<<16,100<<16},
            {180<<16,200<<16},
        };
        drawPolygonTest(test2,4,0x40000000);
        struct vec2 test3[3] = {
            {10<<16,10<<16},
            {(20<<16)+1,10<<16},
            {(20<<16)+1,20<<16},
        };
        drawPolygonTest(test3,3,0x00400000);
        struct vec2 test4[3] = {
            {(20<<16)-1,10<<16},
            {30<<16,10<<16},
            {(20<<16)-1,20<<16},
        };
        drawPolygonTest(test4,3,0x40000000);
        */

        mt_identity(&cur_mtx);
        mt_rotateX(&cur_mtx,frame<<24);
        mt_translate(&cur_mtx,-3<<16,0,0);
        mt_translate(&cur_mtx,-eyeX,-eyeY,-eyeZ);
        draw_model(cube_verts,sizeof(cube_verts)/sizeof(cube_verts[0]),cube_faces,4);

        mt_identity(&cur_mtx);
        mt_rotateY(&cur_mtx,frame<<24);
        mt_translate(&cur_mtx,-eyeX,-eyeY,-eyeZ);
        draw_model(cube_verts,sizeof(cube_verts)/sizeof(cube_verts[0]),cube_faces,4);

        mt_identity(&cur_mtx);
        mt_rotateZ(&cur_mtx,frame<<24);
        mt_translate(&cur_mtx,+3<<16,0,0);
        mt_translate(&cur_mtx,-eyeX,-eyeY,-eyeZ);
        draw_model(cube_verts,sizeof(cube_verts)/sizeof(cube_verts[0]),cube_faces,4);



        //printf("Before submit \n");
        HAL_submit_frame(&screen);
        //printf("After submit \n");
        HAL_flip_buffers();
        //for(;;){}
        frame++;
    }
}


